/**
 * EasyBeans
 * Copyright (C) 2006 Bull S.A.S.
 * Contact: easybeans@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: SessionFactory.java 871 2006-07-13 14:23:56Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.objectweb.easybeans.container.session;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

import javax.ejb.ApplicationException;
import javax.ejb.NoSuchEJBException;

import org.objectweb.easybeans.api.EZBContainer;
import org.objectweb.easybeans.api.FactoryException;
import org.objectweb.easybeans.api.bean.EasyBeansSB;
import org.objectweb.easybeans.api.bean.IBeanInfo;
import org.objectweb.easybeans.container.AbsFactory;
import org.objectweb.easybeans.container.info.SessionBeanInfo;
import org.objectweb.easybeans.log.JLog;
import org.objectweb.easybeans.log.JLogFactory;
import org.objectweb.easybeans.pool.PoolException;
import org.objectweb.easybeans.rpc.JEJBResponse;
import org.objectweb.easybeans.rpc.api.EJBRequest;
import org.objectweb.easybeans.rpc.api.EJBResponse;
import org.objectweb.easybeans.rpc.api.RPCException;

/**
 * This class manages the session bean and its creation/lifecycle.
 * @param <PoolType> the type of bean instance.
 * @author Florent Benoit
 */
public abstract class SessionFactory<PoolType extends EasyBeansSB> extends AbsFactory<PoolType> {

    /**
     * Logger.
     */
    private static JLog logger = JLogFactory.getLog(SessionFactory.class);

    /**
     * Session Desc (deployment info).
     */
    private SessionBeanInfo sessionBeanInfo = null;

    /**
     * Builds a new factory with a given name and its container.
     * @param className name of this factory (name of class that is managed)
     * @param container the root component of this factory.
     * @throws FactoryException if class can't be loaded.
     */
    public SessionFactory(final String className, final EZBContainer container) throws FactoryException {
        super(className, container);
    }

    /**
     * Stops the factory.
     */
    public void stop() {
        // remove pool
        try {
            getPool().stop();
        } catch (PoolException e) {
            logger.error("Problem when stopping the factory", e);
        }

    }

    /**
     * @return information of the current bean.
     */
    public IBeanInfo getBeanInfo() {
        return sessionBeanInfo;
    }

    /**
     * @return information of the current bean.
     */
    public SessionBeanInfo getSessionBeanInfo() {
        return sessionBeanInfo;
    }

    /**
     * Sets the information object for a session bean.
     * @param sessionBeanInfo information on the bean.
     */
    public void setSessionBeanInfo(final SessionBeanInfo sessionBeanInfo) {
        this.sessionBeanInfo = sessionBeanInfo;
    }

    /**
     * Gets a new ID or a null value.
     * @param beanId given id.
     * @return new id
     */
    protected abstract Long getId(final Long beanId);

    /**
     * Creates an instance with the given hint.
     * @param clue a clue given by the Pool. Could be null.
     * @throws PoolException if instance cannot be created.
     * @return the created instance.
     */
    public PoolType create(final Long clue) throws PoolException {
        PoolType instance = null;
        try {
            instance = getBeanClass().newInstance();
        } catch (InstantiationException e) {
            throw new PoolException("Cannot create a new instance", e);
        } catch (IllegalAccessException e) {
            throw new PoolException("Cannot create a new instance", e);
        }

        // Set the factory
        instance.setEasyBeansFactory(this);

        // Init the sessionContext
        EasyBeansSessionContext<PoolType> sessionContext = new EasyBeansSessionContext<PoolType>(instance);
        instance.setEasyBeansSessionContext(sessionContext);

        ClassLoader oldClassLoader = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(getContainer().getClassLoader());
        try {
            // Call injection
            injectResources(instance);

            // post construct callback
            instance.postConstructEasyBeansLifeCycle();
        } finally {
            Thread.currentThread().setContextClassLoader(oldClassLoader);
        }

        return instance;
    }

    /**
     * A request comes to the bean factory and needs to be handled.<br>
     * A response is done which contains the answer.
     * @param request the EJB request.
     * @return a response that have been processed by the factory.
     */
    @Override
    public EJBResponse rpcInvoke(final EJBRequest request) {
        // get hash of the request
        long hash = request.getMethodHash();

        // Get Args (use context classloader for the Serialization)
        Object[] args = null;
        ClassLoader oldClassLoader = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(getContainer().getClassLoader());
        try {
            args = request.getMethodArgs();
        } finally {
            Thread.currentThread().setContextClassLoader(oldClassLoader);
        }

        // call the method
        return localCall(hash, args, request.getBeanId());

    }

    /**
     * Gets a bean for the given id.
     * @param beanId id of the expected bean.
     * @return a Stateless bean.
     * @throws IllegalArgumentException if bean is not found.
     */
    protected abstract PoolType getBean(final Long beanId) throws IllegalArgumentException;

    /**
     * Do a local call on a method of this factory.
     * @param hash the hash of the method to execute.
     * @param methodArgs the arguments of the method
     * @param beanId the id of the bean that we want (stateful).
     * @return response container new id (if any) and value.
     */
    public EJBResponse localCall(final long hash, final Object[] methodArgs, final Long beanId) {
        Long id = getId(beanId);

        // build EJB Response and set the id
        EJBResponse ejbResponse = new JEJBResponse();
        ejbResponse.setBeanId(id);

        PoolType bean = null;
        try {
            bean = getBean(id);
        } catch (IllegalArgumentException e) {
            ejbResponse.setRPCException(new RPCException("Cannot get element in the pool", e));
            return ejbResponse;
        } catch (NoSuchEJBException e) {
            ejbResponse.setRPCException(new RPCException("Bean has been removed", e));
            return ejbResponse;
        }

        Method m = getHashes().get(Long.valueOf(hash));

        if (m == null) {
            ejbResponse.setRPCException(new RPCException("Cannot find method called on the bean '" + getClassName() + "'."));
            return ejbResponse;
        }

        Object value = null;

        // set ClassLoader
        ClassLoader oldClassLoader = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(getContainer().getClassLoader());
        try {
            value = m.invoke(bean, methodArgs);
        } catch (IllegalArgumentException e) {
            ejbResponse.setRPCException(new RPCException(e));
        } catch (IllegalAccessException e) {
            ejbResponse.setRPCException(new RPCException(e));
        } catch (InvocationTargetException e) {
            Throwable cause = e.getCause();
            RPCException rpcException = new RPCException(cause);
            // ApplicationException ?
            ApplicationException applicationException = getBeanInfo().getApplicationExceptions().get(cause.getClass().getName());
            if (applicationException != null) {
                rpcException.setApplicationException();
            }
            ejbResponse.setRPCException(rpcException);
        } finally {
            Thread.currentThread().setContextClassLoader(oldClassLoader);
            // push back into the pool
            try {
                getPool().release(bean);
            } catch (PoolException e) {
                ejbResponse.setRPCException(new RPCException("cannot release bean", e));
            }
        }
        ejbResponse.setValue(value);
        return ejbResponse;

    }

}
