/**
 * EasyBeans
 * Copyright (C) 2006 Bull S.A.S.
 * Contact: easybeans@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: ScanCommonVisitor.java 377 2006-04-21 08:22:46Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.objectweb.easybeans.deployment.annotations.analyzer;

import java.util.HashMap;
import java.util.Map;

import org.objectweb.asm.AnnotationVisitor;
import org.objectweb.asm.commons.EmptyVisitor;
import org.objectweb.easybeans.deployment.annotations.metadata.interfaces.ISharedMetadata;

/**
 * This classes analyses annotation (could be class, method, attribute, etc).
 * It is extended by classes for a specific type Class, Method, Attribute.
 * @param <T> the kind of metadata visited by this visitor.
 * @author Florent Benoit
 */
public abstract class ScanCommonVisitor<T extends ISharedMetadata> extends EmptyVisitor {

    /**
     * Annotation visitor which do nothing.
     */
    private EmptyVisitor emptyVisitor = null;

    /**
     * Map of annotation visitors used by this visitor.<br>
     * The key is the annotation descriptor (asm)
     */
    private Map<String, AnnotationVisitor> annotationVisitors = null;

    /**
     * Constructor
     * Build map of visitors.
     */
    public ScanCommonVisitor() {
        annotationVisitors = new HashMap<String, AnnotationVisitor>();
    }

    /**
     * Build visitors used by this one.
     * @param annotationMetadata metadata used
     */
    protected void initVisitors(final T annotationMetadata) {
        emptyVisitor = new EmptyVisitor();

        // add @EJB
        annotationVisitors.put(JavaxEjbEJBVisitor.TYPE,
                new JavaxEjbEJBVisitor<T>(annotationMetadata));

        // add @Resource
        annotationVisitors.put(JavaxAnnotationResourceVisitor.TYPE,
                new JavaxAnnotationResourceVisitor<T>(annotationMetadata));


        // add @PersistenceContext
        annotationVisitors.put(JavaxPersistencePersistenceContextVisitor.TYPE,
                new JavaxPersistencePersistenceContextVisitor<T>(annotationMetadata));

        // add @PersistenceUnit
        annotationVisitors.put(JavaxPersistencePersistenceUnitVisitor.TYPE,
                new JavaxPersistencePersistenceUnitVisitor<T>(annotationMetadata));

    }


    /**
     * @return empty visitor
     */
    protected EmptyVisitor getEmptyVisitor() {
        return emptyVisitor;
    }

    /**
     * Visits an annotation of the class.
     * @param desc the class descriptor of the annotation class.
     * @param visible <tt>true</tt> if the annotation is visible at runtime.
     * @return a non null visitor to visit the annotation values.
     */
    @Override
    public AnnotationVisitor visitAnnotation(final String desc, final boolean visible) {
        AnnotationVisitor av = annotationVisitors.get(desc);
        if (av != null) {
            return av;
        }
        return emptyVisitor;
    }

    /**
     * @return the visitors used by this scanner.
     */
    protected Map<String, AnnotationVisitor> getAnnotationVisitors() {
        return annotationVisitors;
    }


}
