/**
 * EasyBeans
 * Copyright (C) 2006 Bull S.A.S.
 * Contact: easybeans@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: JNDIResolver.java 242 2006-03-22 10:32:44Z benoitf $
 * --------------------------------------------------------------------------
 */
package org.objectweb.easybeans.deployment.resolver;

import java.util.HashMap;
import java.util.Map;

import org.objectweb.easybeans.container.JContainer3;
import org.objectweb.easybeans.deployment.Deployment;
import org.objectweb.easybeans.deployment.annotations.analyzer.AnnotationDeploymentAnalyzer;
import org.objectweb.easybeans.deployment.annotations.impl.JCommonBean;
import org.objectweb.easybeans.deployment.annotations.impl.JLocal;
import org.objectweb.easybeans.deployment.annotations.impl.JRemote;
import org.objectweb.easybeans.deployment.annotations.metadata.ClassAnnotationMetadata;
import org.objectweb.easybeans.deployment.annotations.metadata.EjbJarAnnotationMetadata;
import org.objectweb.easybeans.log.JLog;
import org.objectweb.easybeans.log.JLogFactory;

/**
 * Classes used to resolve JNDI Name for a given deployment.
 * @author Florent Benoit
 */
public class JNDIResolver {

    /**
     * Logger.
     */
    private JLog logger = JLogFactory.getLog(JNDIResolver.class);

    /**
     * Name.
     */
    public static final String NAME = "jndi.resolver";

    /**
     * Map with key = interface name, value = jndi name.
     */
    private Map<String, String> interfaces = null;

    /**
     * Map with key = bean name, value = &lt;interface name, jndi name&gt;.
     */
    private Map<String, Map<String, String>> beans = null;

    /**
     * Constructor.
     */
    public JNDIResolver() {
        interfaces = new HashMap<String, String>();
        beans = new HashMap<String, Map<String, String>>();
    }

    /**
     * Build a new resolver with the given deployment.
     * @param deployment given deployment object to use.
     */
    public JNDIResolver(final Deployment deployment) {
        this();
        addDeployment(deployment);
    }

    /**
     * Add a given deployment object to this resolver.
     * @param deployment to add to the resolver.
     */
    public void addDeployment(final Deployment deployment) {
        // Get Analyzer
        AnnotationDeploymentAnalyzer analyzer = deployment.getAnnotationDeploymentAnalyzer();

        // Get metadata for jar file analyzed
        EjbJarAnnotationMetadata ejbJarAnnotationMetadata = analyzer.getEjbJarAnnotationMetadata();
        addEjbJarAnnotationMetadata(ejbJarAnnotationMetadata);
    }

    /**
     * Adds the given metadata to the resolver.
     * @param ejbJarAnnotationMetadata the metadata for a given jar file
     */
    public void addEjbJarAnnotationMetadata(final EjbJarAnnotationMetadata ejbJarAnnotationMetadata) {
        // For each bean, get the interfaces and add the jndiName mapping.
        for (ClassAnnotationMetadata classAnnotationMetadata : ejbJarAnnotationMetadata
                .getClassAnnotationMetadataCollection()) {
            if (classAnnotationMetadata.isBean()) {

                // Also, go through the EJB annotations (using ejb name)
                JCommonBean jCommonBean = classAnnotationMetadata.getJCommonBean();
                String mappedName = null;
                String beanName = null;
                if (jCommonBean != null) {
                    // jndi name ?
                    mappedName = jCommonBean.getMappedName();
                    // bean name ?
                    beanName = jCommonBean.getName();

                }

                // SessionBusinessInterfaceFinder.resolve(classAnnotationMetadata);
                JLocal localItfs = classAnnotationMetadata.getLocalInterfaces();
                JRemote remoteItfs = classAnnotationMetadata.getRemoteInterfaces();
                if (localItfs != null) {
                    for (String itf : localItfs.getInterfaces()) {
                        addInterface(itf, classAnnotationMetadata.getClassName(), "Local", mappedName, beanName);
                    }
                }
                if (remoteItfs != null) {
                    for (String itf : remoteItfs.getInterfaces()) {
                        addInterface(itf, classAnnotationMetadata.getClassName(), "Remote", mappedName, beanName);
                    }
                }
            }
        }
    }

    /**
     * Add the jndi name for a given interface and a given bean name (may be
     * null).
     * @param itf the interface of the bean
     * @param beanClassName the class of the bean
     * @param mode local/remote
     * @param mappedName the mappedName (could be used as JNDI name)
     * @param beanName the name of the bean
     */
    private void addInterface(final String itf, final String beanClassName, final String mode, final String mappedName,
            final String beanName) {
        String jndiName = JContainer3.jndiNameEncode(beanClassName, itf, mode);
        if (mappedName != null) {
            jndiName = mappedName;
        }
        String itfName = itf.replace("/", ".");
        interfaces.put(itfName, jndiName);

        if (beanName != null && !beanName.equals("")) {
            Map<String, String> tmpBeanMap = beans.get(beanName);
            if (tmpBeanMap == null) {
                tmpBeanMap = new HashMap<String, String>();
                beans.put(beanName, tmpBeanMap);
            }
            tmpBeanMap.put(itfName, jndiName);
            beans.put(beanName, tmpBeanMap);
            logger.debug("Adding jndi name {0} for beanName {1} with itf {2} and className {3}", jndiName, beanName,
                    itfName, beanClassName);
        } else {
            logger.debug("Adding jndi name {0} with itf {1} and className {2}", jndiName, itfName, beanClassName);
        }
    }

    /**
     * Gets jndi name for a given interface and a bean name.
     * @param itf the name of the interface.
     * @param beanName the name of the bean.
     * @return the jndi name or null if not found.
     */
    public String getJndiNameInterface(final String itf, final String beanName) {
        if (beanName != null && !beanName.equals("")) {
            Map<String, String> lst = beans.get(beanName);
            if (lst != null) {
                return lst.get(itf);
            }
            throw new IllegalStateException("No bean with name '" + beanName + "' was found.");

        }
        return interfaces.get(itf);
    }

    /**
     * @return string representation of this object
     */
    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder("JNDIResolver[");
        sb.append("Interfaces =[");
        sb.append(interfaces);
        sb.append("], Beans =");
        sb.append(beans);
        sb.append("]]");

        return sb.toString();
    }
}
