/**
 * EasyBeans
 * Copyright (C) 2006 Bull S.A.S.
 * Contact: easybeans@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: InjectionClassAdapter.java 793 2006-06-27 14:03:51Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.objectweb.easybeans.enhancer.injection;

import static org.objectweb.easybeans.injection.JNDILookupHelper.JndiType.JAVA_COMP;
import static org.objectweb.easybeans.injection.JNDILookupHelper.JndiType.JAVA_COMP_ENV;
import static org.objectweb.easybeans.injection.JNDILookupHelper.JndiType.REGISTRY;

import java.util.Arrays;
import java.util.List;
import java.util.Map;

import javax.ejb.SessionContext;
import javax.ejb.TimerService;
import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.sql.DataSource;
import javax.transaction.UserTransaction;

import org.objectweb.asm.ClassAdapter;
import org.objectweb.asm.ClassVisitor;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Opcodes;
import org.objectweb.asm.Type;
import org.objectweb.easybeans.api.Factory;
import org.objectweb.easybeans.container.session.EasyBeansSessionContext;
import org.objectweb.easybeans.deployment.annotations.JMethod;
import org.objectweb.easybeans.deployment.annotations.impl.JAnnotationResource;
import org.objectweb.easybeans.deployment.annotations.impl.JEjbEJB;
import org.objectweb.easybeans.deployment.annotations.impl.JavaxPersistenceContext;
import org.objectweb.easybeans.deployment.annotations.impl.JavaxPersistenceUnit;
import org.objectweb.easybeans.deployment.annotations.metadata.ClassAnnotationMetadata;
import org.objectweb.easybeans.deployment.annotations.metadata.EjbJarAnnotationMetadata;
import org.objectweb.easybeans.deployment.annotations.metadata.FieldAnnotationMetadata;
import org.objectweb.easybeans.deployment.annotations.metadata.MethodAnnotationMetadata;
import org.objectweb.easybeans.deployment.resolver.JNDIResolver;
import org.objectweb.easybeans.enhancer.CommonClassGenerator;
import org.objectweb.easybeans.enhancer.interceptors.EasyBeansInvocationContextGenerator;
import org.objectweb.easybeans.enhancer.lib.MethodRenamer;
import org.objectweb.easybeans.injection.JNDILookupHelper.JndiType;
import org.objectweb.easybeans.log.JLog;
import org.objectweb.easybeans.log.JLogFactory;
import org.omg.CORBA.ORB;

/**
 * This class adds methods which will inject resources in the bean class.
 * @author Florent Benoit
 */
public class InjectionClassAdapter extends ClassAdapter implements Opcodes {

    /**
     * Logger.
     */
    private static JLog logger = JLogFactory.getLog(InjectionClassAdapter.class);

    /**
     * Metadata available by this adapter for a class.
     */
    private ClassAnnotationMetadata classAnnotationMetadata;

    /**
     * Map containing informations for enhancers.
     */
    private Map<String, Object> map = null;

    /**
     * Is that generated method is static (client case).
     */
    private boolean staticMode = false;

    /**
     * javax.ejb.SessionContext interface.
     */
    private static final String SESSION_CONTEXT = SessionContext.class.getName();

    /**
     * org.omg.CORBA.ORB interface.
     */
    private static final String ORB_ITF = ORB.class.getName();

    /**
     * javax.sql.DataSource interface.
     */
    private static final String DATASOURCE_ITF = DataSource.class.getName();

    /**
     * javax.transaction.UserTransaction interface.
     */
    private static final String USERTRANSACTION_ITF = UserTransaction.class.getName();

    /**
     * Entity Manager interface.
     */
    private static final String ENTITYMANAGER_ITF = EntityManager.class.getName();

    /**
     * Entity Manager Factory interface.
     */
    private static final String ENTITYMANAGERFACTORY_ITF = EntityManagerFactory.class.getName();

    /**
     * javax.ejb.TimerService interface.
     */
    private static final String TIMERSERVICE_ITF = TimerService.class.getName();

    /**
     * EasyBeansSessionContext type descriptor.
     */
    private static final String EZB_SESSIONCONTEXT_DESC  = Type.getDescriptor(EasyBeansSessionContext.class);

    /**
     * Defines java.lang.Object class.
     */
    public static final String JAVA_LANG_OBJECT = "java/lang/Object";

    /**
     * Injected method name.
     */
    public static final String INJECTED_METHOD = "injectedByEasyBeans";

    /**
     * JMethod object for injectedByEasyBeans.
     */
    public static final JMethod INJECTED_JMETHOD = new JMethod(ACC_PUBLIC, MethodRenamer.encode(INJECTED_METHOD), "()V", null,
            new String[] {"org/objectweb/easybeans/api/injection/EasyBeansInjectionException"});

    /**
     * List of injected methods.
     */
    public static final String[] INJECTED_METHODS = new String[] {"getEasyBeansSessionContext", "setEasyBeansSessionContext",
            "getEasyBeansFactory", "setEasyBeansFactory"};

    /**
     * Replace length to create default JNDI names.
     */
    private static final int LENGTH = 3;

    /**
     * Constructor.
     * @param classAnnotationMetadata object containing all attributes of the
     *        class
     * @param cv the class visitor to which this adapter must delegate calls.
     * @param map a map allowing to give some objects to the adapter.
     * @param staticMode - Is that generated method is static (client case).
     */
    public InjectionClassAdapter(final ClassAnnotationMetadata classAnnotationMetadata, final ClassVisitor cv,
            final Map<String, Object> map, final boolean staticMode) {
        super(cv);
        this.classAnnotationMetadata = classAnnotationMetadata;
        this.map = map;
        this.staticMode = staticMode;
    }

    /**
     * Visits the end of the class. This method, which is the last one to be
     * called, is used to inform the visitor that all the fields and methods of
     * the class have been visited.
     */
    @Override
    public void visitEnd() {
        super.visitEnd();

        // now, adds the injected method
        addInjectedMethod();

        // Adds methods if it's not a bean (as it should have been already added by the bean class adapter).
        // If it's a super class of a bean, as the class that will be instantiated
        // is the bean's class, the methods won't be used.
        // It it's an interceptor class, the interceptor manager will call the setters methods
        if (!classAnnotationMetadata.isBean()) {
            addDefaultMethods();
        }
    }

    /**
     * Generated methods allowing to set a context and a factory.
     * This allows to set on injectors the bean's session context and its factory.
     */
    private void addDefaultMethods() {
        // Adds the factory attribute and its getter/setter.
        CommonClassGenerator.addFieldGettersSetters(cv, classAnnotationMetadata.getClassName(), "easyBeansFactory",
                Factory.class);

        // Adds the sessionContext attribute and its getter/setter.
        CommonClassGenerator.addFieldGettersSetters(cv, classAnnotationMetadata.getClassName(),
                "easyBeansSessionContext", EasyBeansSessionContext.class);
    }

    /**
     * Generates the injectedByEasyBeans() method on the current class.
     */
    private void addInjectedMethod() {
        int access = ACC_PUBLIC;
        if (staticMode) {
            access = access + ACC_STATIC;
        }

        MethodVisitor mv = cv.visitMethod(access, INJECTED_METHOD, "()V", null,
                new String[] {"org/objectweb/easybeans/api/injection/EasyBeansInjectionException"});
        mv.visitCode();

        // First, call the super class method (if the super class has been
        // analyzed) and if there is one
        String superNameClass = classAnnotationMetadata.getSuperName();
        if (superNameClass != null && !superNameClass.equals(JAVA_LANG_OBJECT)) {
            EjbJarAnnotationMetadata jarMetadata = classAnnotationMetadata.getEjbJarAnnotationMetadata();
            ClassAnnotationMetadata superMetadata = jarMetadata.getClassAnnotationMetadata(superNameClass);
            if (superMetadata != null) {
                if (!staticMode) {
                    // generate call to super method : super.INJECTED_METHOD();
                    mv.visitVarInsn(ALOAD, 0);
                    mv.visitMethodInsn(INVOKESPECIAL, superMetadata.getClassName(), INJECTED_METHOD, "()V");
                } else {
                    mv.visitMethodInsn(INVOKESTATIC, superMetadata.getClassName(), INJECTED_METHOD, "()V");
                }
            }
        }

        // If it is a bean, call the interceptorManager and the attributes (like session context and factory)
        if (classAnnotationMetadata.isBean()) {
            String clNameManager = classAnnotationMetadata.getClassName()
            + EasyBeansInvocationContextGenerator.SUFFIX_INTERCEPTOR_MANAGER;

            // this.interceptorManager.setEasyBeansSessionContext(easyBeansSessionContext);
            mv.visitVarInsn(ALOAD, 0);
            mv.visitFieldInsn(GETFIELD, classAnnotationMetadata.getClassName(), "easyBeansInterceptorManager", "L"
                    + clNameManager + ";");
            mv.visitVarInsn(ALOAD, 0);
            mv.visitFieldInsn(GETFIELD, classAnnotationMetadata.getClassName(), "easyBeansSessionContext",
                    "Lorg/objectweb/easybeans/container/session/EasyBeansSessionContext;");
            mv.visitMethodInsn(INVOKEVIRTUAL, clNameManager, "setEasyBeansSessionContext",
                    "(Lorg/objectweb/easybeans/container/session/EasyBeansSessionContext;)V");


            // this.interceptorManager.injectedByEasyBeans();
            mv.visitVarInsn(ALOAD, 0);
            mv.visitFieldInsn(GETFIELD, classAnnotationMetadata.getClassName(), "easyBeansInterceptorManager", "L"
                    + clNameManager + ";");
            mv.visitMethodInsn(INVOKEVIRTUAL, clNameManager, "injectedByEasyBeans", "()V");

        }

        generateBodyInjectedMethod(mv);

        mv.visitInsn(RETURN);
        mv.visitMaxs(0, 0);
        mv.visitEnd();
    }

    /**
     * Generates the body of the injectedByEasyBeans() method if any.<br> Else,
     * do nothing.
     * @param mv the method visitor object used to add some code.
     */
    private void generateBodyInjectedMethod(final MethodVisitor mv) {

        // generates injection for annotations on the class itself
        generateClassInjection(mv);

        // Generates injection for attributes
        generateAttributesInjection(mv);

        // Generates injection for setters methods
        generateSettersInjection(mv);

    }

    /**
     * Generates the calls to populate ENC environment by using annotations on the class itself.
     * @param mv the method visitor used to inject bytecode.
     */
    private void generateClassInjection(final MethodVisitor mv) {
        // Get annotations on the class

        // &#64;PersistenceContexts annotation
        List<JavaxPersistenceContext> javaxPersistencePersistenceContexts = classAnnotationMetadata
                .getJavaxPersistencePersistenceContexts();
        if (javaxPersistencePersistenceContexts != null && javaxPersistencePersistenceContexts.size() > 0) {
            // For each javaxPersistenceContext
            for (JavaxPersistenceContext javaxPersistenceContext : javaxPersistencePersistenceContexts) {
                bindClassPersistenceContext(javaxPersistenceContext, mv);
            }
        }
        // &#64;PersistenceContext annotation
        if (classAnnotationMetadata.isPersistenceContext()) {
            bindClassPersistenceContext(classAnnotationMetadata.getJavaxPersistenceContext(), mv);
        }

        // &#64;PersistenceUnits annotation
        List<JavaxPersistenceUnit> javaxPersistencePersistenceUnits = classAnnotationMetadata
                .getJavaxPersistencePersistenceUnits();
        if (javaxPersistencePersistenceUnits != null && javaxPersistencePersistenceUnits.size() > 0) {
            // For each javaxPersistenceUnit
            for (JavaxPersistenceUnit javaxPersistenceUnit : javaxPersistencePersistenceUnits) {
                bindClassPersistenceUnit(javaxPersistenceUnit, mv);
            }
        }
        // &#64;PersistenceUnit annotation
        if (classAnnotationMetadata.isPersistenceUnit()) {
            bindClassPersistenceUnit(classAnnotationMetadata.getJavaxPersistenceUnit(), mv);
        }

        // &#64;EJBs annotation
        List<JEjbEJB> jEjbs = classAnnotationMetadata.getJEjbEJBs();
        if (jEjbs != null && jEjbs.size() > 0) {
            // For each jEJB
            for (JEjbEJB jEJB : jEjbs) {
                bindClassEJB(jEJB, mv);
            }
        }
        // &#64;EJB annotation
        JEjbEJB jEJB = classAnnotationMetadata.getJEjbEJB();
        if (jEJB != null) {
            // For each ejb, do :
            bindClassEJB(jEJB, mv);
        }


        // &#64;Resources annotation
        List<JAnnotationResource> jAnnotationResources = classAnnotationMetadata.getJAnnotationResources();
        if (jAnnotationResources != null && jAnnotationResources.size() > 0) {
            // For each jAnnotationResource
            for (JAnnotationResource jAnnotationResource : jAnnotationResources) {
                bindResource(jAnnotationResource, mv);
            }
        }
        // &#64;Resource annotation
        JAnnotationResource jAnnotationResource = classAnnotationMetadata.getJAnnotationResource();
        if (jAnnotationResource != null) {
            bindResource(jAnnotationResource, mv);
        }


    }


    /**
     * Generates the calls to methods that will set the attributes value.
     * @param mv the method visitor used to inject bytecode.
     */
    private void generateAttributesInjection(final MethodVisitor mv) {

        for (FieldAnnotationMetadata fieldMetaData : classAnnotationMetadata.getFieldAnnotationMetadataCollection()) {

            // Get type of interface
            Type typeInterface = Type.getType(fieldMetaData.getJField().getDescriptor());
            String itfName = typeInterface.getClassName();

            // &#64;PersistenceContext annotation
            if (fieldMetaData.isPersistenceContext()) {
                // validate
                validateAccessFieldAnnotation(fieldMetaData);

                // Check that attribute is EntityManager
                if (!ENTITYMANAGER_ITF.equals(itfName)) {
                    throw new IllegalStateException(
                            "Trying to applied @PersistenceContext on an invalid field in the class '"
                                    + classAnnotationMetadata.getClassName() + "', field = " + fieldMetaData);
                }
                logger.debug("Add injection for PersistenceContext on attribute {0} of class {1}", fieldMetaData
                        .getFieldName(), classAnnotationMetadata.getClassName());
                // add this.em =
                // EntityManagerHelper.getEntityManager(getEasyContext(),
                // "myUnitName", PersistenceContextType.EXTENDED);
                mv.visitVarInsn(ALOAD, 0);

                // call em helper
                addCallEntityManagerHelper(fieldMetaData.getJavaxPersistenceContext(), mv);

                // Set result in the field
                mv.visitFieldInsn(PUTFIELD, classAnnotationMetadata.getClassName(), fieldMetaData.getFieldName(),
                        "Ljavax/persistence/EntityManager;");
            }

            // &#64;PersistenceUnit annotation
            if (fieldMetaData.isPersistenceUnit()) {
                // validate
                validateAccessFieldAnnotation(fieldMetaData);

                // Check that attribute is EntityManager
                if (!ENTITYMANAGERFACTORY_ITF.equals(itfName)) {
                    throw new IllegalStateException(
                            "Trying to applied @PersistenceUnit on an invalid field in the class '"
                                    + classAnnotationMetadata.getClassName() + "', field = " + fieldMetaData);
                }
                logger.debug("Add injection for PersistenceUnit on attribute {0} of class {1}", fieldMetaData
                        .getFieldName(), classAnnotationMetadata.getClassName());
                // add this.emf = EntityManagerHelper.getEntityManagerFactory(getEasyContext(), "myUnitName");

                mv.visitVarInsn(ALOAD, 0);
                // get EMF
                addCallEntityManagerFactoryHelper(fieldMetaData.getJavaxPersistenceUnit(), mv);
                // set attribute
                mv.visitFieldInsn(PUTFIELD, classAnnotationMetadata.getClassName(), fieldMetaData.getFieldName(),
                        "Ljavax/persistence/EntityManagerFactory;");
            }

            // &#64;EJB annotation
            JEjbEJB jEjb = fieldMetaData.getJEjbEJB();
            if (jEjb != null) {
                // validate
                validateAccessFieldAnnotation(fieldMetaData);

                logger.debug("Add injection for EJB on attribute {0} of class {1}", fieldMetaData.getFieldName(),
                        classAnnotationMetadata.getClassName());

                JNDIResolver jndiResolver = (JNDIResolver) map.get(JNDIResolver.NAME);
                // ejbName ?
                String beanName = jEjb.getBeanName();
                String jndiName = jndiResolver.getJndiNameInterface(itfName, beanName);

                if (jndiName == null) {
                    logger.error("No jndi name found on class {0} for interface {1} and beanName {2}",
                            classAnnotationMetadata.getClassName(), itfName, beanName);
                } else {
                    logger.debug("Asking jndi name on class {0} for interface {1} and beanName {2}. Result = {3}",
                            classAnnotationMetadata.getClassName(), itfName, beanName, jndiName);
                    callAttributeJndi(jndiName, typeInterface, mv, fieldMetaData, classAnnotationMetadata
                            .getClassName(), REGISTRY);
                    callBindAttributeJndi(jEjb.getName(), jndiName, mv, fieldMetaData);
                }
            }

            // &#64;Resource annotation
            JAnnotationResource jAnnotationResource = fieldMetaData.getJAnnotationResource();
            if (jAnnotationResource != null) {
                String mappedName = jAnnotationResource.getMappedName();

                // validate
                validateAccessFieldAnnotation(fieldMetaData);

                if (SESSION_CONTEXT.equals(itfName)) {
                    logger.debug("Add injection for @Resource on attribute {0} of class {1} for the type {2}",
                            fieldMetaData.getFieldName(), classAnnotationMetadata.getClassName(), itfName);

                    // this.attribute = getEasyBeansSessionContext();
                    mv.visitVarInsn(ALOAD, 0);
                    addCallGetEasyBeansContext(mv);
                    mv.visitFieldInsn(PUTFIELD, classAnnotationMetadata.getClassName(), fieldMetaData.getFieldName(),
                            "Ljavax/ejb/SessionContext;");
                    // Define the type (if missing)
                    jAnnotationResource.setType(SESSION_CONTEXT);

                    // Set default name if not present.
                    jAnnotationResource.setName(getJndiName(jAnnotationResource.getName(), fieldMetaData));

                    bindResource(jAnnotationResource, mv);

                } else if (isEnvEntry(typeInterface)) { // Env-Entry
                    callAttributeJndi(jAnnotationResource.getName(), typeInterface, mv, fieldMetaData,
                            classAnnotationMetadata.getClassName(), JAVA_COMP_ENV);
                } else if (DATASOURCE_ITF.equals(itfName)) {
                    if (mappedName != null && !mappedName.equals("")) {
                        callAttributeJndi(mappedName, typeInterface, mv, fieldMetaData,
                                classAnnotationMetadata.getClassName(), REGISTRY);
                        callBindAttributeJndi(jAnnotationResource.getName(), mappedName, mv, fieldMetaData);
                    } else {
                    //TODO: Change from REGISTRY to JAVA_COMP_ENV
                    callAttributeJndi(jAnnotationResource.getName(), typeInterface, mv, fieldMetaData,
                            classAnnotationMetadata.getClassName(), REGISTRY);
                    }
                } else if (USERTRANSACTION_ITF.equals(itfName)) {
                    callAttributeJndi("UserTransaction", typeInterface, mv, fieldMetaData,
                            classAnnotationMetadata.getClassName(), JAVA_COMP);
                } else if (TIMERSERVICE_ITF.equals(itfName)) {
                    // Needs to get timerservice with the sessioncontext.
                    //this.fieldtimerService = getEasyBeansSessionContext().getTimerService();
                    mv.visitVarInsn(ALOAD, 0);
                    addCallGetEasyBeansContext(mv);
                    mv.visitMethodInsn(INVOKEVIRTUAL,
                            "org/objectweb/easybeans/container/session/EasyBeansSessionContext", "getTimerService",
                            "()Ljavax/ejb/TimerService;");
                    mv.visitFieldInsn(PUTFIELD, classAnnotationMetadata.getClassName(), fieldMetaData.getFieldName(),
                            "Ljavax/ejb/TimerService;");
                } else if (ORB_ITF.equals(itfName)) {
                    // this.orb = ORBInitHelper.getORB(); (non static)
                    if (!staticMode) {
                        mv.visitVarInsn(ALOAD, 0);
                    }
                    mv.visitMethodInsn(INVOKESTATIC, "org/objectweb/easybeans/injection/ORBInitHelper", "getORB",
                            "()Lorg/omg/CORBA/ORB;");
                    mv.visitFieldInsn(setField(), classAnnotationMetadata.getClassName(), fieldMetaData.getFieldName(),
                            "Lorg/omg/CORBA/ORB;");

                }

            }

        }
    }

    /**
     * Generates the calls to methods that will call the setters methods.
     * @param mv the method visitor used to inject bytecode.
     */
    private void generateSettersInjection(final MethodVisitor mv) {

        for (MethodAnnotationMetadata methodMetaData : classAnnotationMetadata.getMethodAnnotationMetadataCollection()) {
            // Ignore inherited methods (managed by super class)
            if (methodMetaData.isInherited()) {
                continue;
            }

            JAnnotationResource jAnnotationResource = methodMetaData.getJAnnotationResource();
            // &#64;Resource annotation
            if (jAnnotationResource != null) {
                Type typeInterface = validateSetterMethod(methodMetaData);
                String itfName = typeInterface.getClassName();

                // Env-Entry
                if (isEnvEntry(typeInterface)) {
                    callMethodJndiEnv(jAnnotationResource.getName(), typeInterface, mv, methodMetaData,
                            classAnnotationMetadata.getClassName(), JAVA_COMP_ENV);
                } else if (DATASOURCE_ITF.equals(itfName)) {
                    //TODO: Change from REGISTRY to JAVA_COMP_ENV
                    callMethodJndiEnv(jAnnotationResource.getName(), typeInterface, mv, methodMetaData,
                            classAnnotationMetadata.getClassName(), REGISTRY);
                } else if (USERTRANSACTION_ITF.equals(itfName)) {
                    callMethodJndiEnv("UserTransaction", typeInterface, mv, methodMetaData,
                            classAnnotationMetadata.getClassName(), JAVA_COMP);
                } else if (SESSION_CONTEXT.equals(itfName)) {
                    // add call to : setterMethod(getEasyBeansSessionContext());
                    mv.visitVarInsn(ALOAD, 0);
                    addCallGetEasyBeansContext(mv);
                    mv.visitMethodInsn(INVOKEVIRTUAL, classAnnotationMetadata.getClassName(), methodMetaData.getMethodName(),
                            "(Ljavax/ejb/SessionContext;)V");

                    // Define the type (if missing)
                    jAnnotationResource.setType(SESSION_CONTEXT);
                    // Set default name if not present.
                    jAnnotationResource.setName(getJndiName(jAnnotationResource.getName(), methodMetaData));
                    bindResource(jAnnotationResource, mv);
                }
            }

            // &#64;EJB annotation
            JEjbEJB jEjb = methodMetaData.getJEjbEJB();
            if (jEjb != null) {
                logger.debug("Add injection for EJB on method {0} of class {1}", methodMetaData.getMethodName(),
                        classAnnotationMetadata.getClassName());

                Type typeInterface = validateSetterMethod(methodMetaData);
                String itfName = typeInterface.getClassName();

                JNDIResolver jndiResolver = (JNDIResolver) map.get(JNDIResolver.NAME);
                // ejbName ?
                String beanName = jEjb.getBeanName();
                String jndiName = jndiResolver.getJndiNameInterface(itfName, beanName);
                logger.debug("Asking jndi name on class {0} for interface {1} and beanName {2}. Result = {3}",
                        classAnnotationMetadata.getClassName(), itfName, beanName, jndiName);

                callMethodJndiEnv(jndiName, typeInterface, mv, methodMetaData, classAnnotationMetadata.getClassName(), REGISTRY);

                // get enc name (or the default name) and bind result
                String encName = getJndiName(jEjb.getName(), methodMetaData);
                callBindLookupJndiRef(encName, jndiName, mv);
            }

            // &#64;PersistenceContext annotation
            if (methodMetaData.isPersistenceContext()) {
                Type typeInterface = validateSetterMethod(methodMetaData);
                String itfName = typeInterface.getClassName();

                // Check that arg of the method is EntityManager
                if (!ENTITYMANAGER_ITF.equals(itfName)) {
                    throw new IllegalStateException(
                            "Trying to applied @PersistenceContext on an invalid method in the class '"
                                    + classAnnotationMetadata.getClassName() + "', method = " + methodMetaData);
                }
                logger.debug("Add injection for PersistenceContext on method {0} of class {1}", methodMetaData
                        .getMethodName(), classAnnotationMetadata.getClassName());
                // add
                // setterName(EntityManagerHelper.getEntityManager(getEasyContext(),
                // "myUnitName", PersistenceContextType.EXTENDED);

                mv.visitVarInsn(ALOAD, 0);

                // call em helper
                addCallEntityManagerHelper(methodMetaData.getJavaxPersistenceContext(), mv);

                // call setter method
                mv.visitMethodInsn(INVOKEVIRTUAL, classAnnotationMetadata.getClassName(), methodMetaData
                        .getMethodName(), "(Ljavax/persistence/EntityManager;)V");

            }

            // &#64;PersistenceUnit annotation
            if (methodMetaData.isPersistenceUnit()) {
                Type typeInterface = validateSetterMethod(methodMetaData);
                String itfName = typeInterface.getClassName();
                // Check that attribute is EntityManager
                if (!ENTITYMANAGERFACTORY_ITF.equals(itfName)) {
                    throw new IllegalStateException(
                            "Trying to applied @PersistenceUnit on an invalid method in the class '"
                                    + classAnnotationMetadata.getClassName() + "', method = " + methodMetaData);
                }
                logger.debug("Add injection for PersistenceUnit on on method {0} of class {1}", methodMetaData
                        .getMethodName(), classAnnotationMetadata.getClassName());
                // add
                // setterName(EntityManagerHelper.getEntityManagerFactory(getEasyContext(),
                // "myUnitName"));

                mv.visitVarInsn(ALOAD, 0);
                // get EMF
                addCallEntityManagerFactoryHelper(methodMetaData.getJavaxPersistenceUnit(), mv);
                // call setter method
                mv.visitMethodInsn(INVOKEVIRTUAL, classAnnotationMetadata.getClassName(), methodMetaData
                        .getMethodName(), "(Ljavax/persistence/EntityManagerFactory;)V");
            }

        }
    }


    /**
     * Ensure that this method is a valid setter method and return ASM type of the first arg of the method.
     * @param methodMetaData the metadata to check
     * @return ASM type of the first arg of the method.
     */
    private Type validateSetterMethod(final MethodAnnotationMetadata methodMetaData) {
        // validate access
        validateAccessMethodAnnotation(methodMetaData);

        JMethod jMethod = methodMetaData.getJMethod();
        // Should be a setter
        if (!jMethod.getName().startsWith("set") || jMethod.getName().equalsIgnoreCase("set")) {
            throw new IllegalStateException("Method '" + jMethod
                    + "' is invalid. Should be in the setter form setXXX().");
        }

        // Get type of interface
        // Get arguments of the method.
        Type[] args = Type.getArgumentTypes(jMethod.getDescriptor());
        if (args.length != 1) {
            throw new IllegalStateException("Method args '" + Arrays.asList(args) + "' for method '" + jMethod
                    + "' are invalid. Length should be of 1.");
        }
        return args[0];
    }

    /**
     * Return true if the given type is a type used in env-entry.
     * @param type an ASM type.
     * @return true if this entry is used in env-entry.
     */
    private boolean isEnvEntry(final Type type) {
        String itfName = type.getClassName();
        return String.class.getName().equals(itfName) || Boolean.TYPE.getName().equals(itfName)
                || Byte.TYPE.getName().equals(itfName) || Character.TYPE.getName().equals(itfName)
                || Double.TYPE.getName().equals(itfName) || Float.TYPE.getName().equals(itfName)
                || Integer.TYPE.getName().equals(itfName) || Long.TYPE.getName().equals(itfName)
                || Short.TYPE.getName().equals(itfName);
    }

    /**
     * Add a call to getEasyBeansSessionContext() method in the given method visitor.
     * @param mv the method visitor on which instructions are added
     */
    private void addCallGetEasyBeansContext(final MethodVisitor mv) {
        mv.visitVarInsn(ALOAD, 0);
        mv.visitMethodInsn(INVOKEVIRTUAL, classAnnotationMetadata.getClassName(),
                "getEasyBeansSessionContext", "()" + EZB_SESSIONCONTEXT_DESC);
    }

    /**
     * Add a call to EntityManagerHelper class (PersistenceContext) :
     * EntityManagerHelper.getEntityManager(getEasyBeansSessionContext(),
     * unitName, type of persistence).
     * @param javaxPersistenceContext informations on the persistence context
     * @param mv the method visitor on which instructions are added
     */
    private void addCallEntityManagerHelper(final JavaxPersistenceContext javaxPersistenceContext, final MethodVisitor mv) {
        // get EasyBeansSessionContext
        addCallGetEasyBeansContext(mv);

        // Persistence-unit name
        mv.visitLdcInsn(javaxPersistenceContext.getUnitName());

        // Transaction Type
        mv.visitFieldInsn(GETSTATIC, "javax/persistence/PersistenceContextType", javaxPersistenceContext.getType().toString(),
                "Ljavax/persistence/PersistenceContextType;");
        // Call EntityManagerHelper
        mv
        .visitMethodInsn(
                INVOKESTATIC,
                "org/objectweb/easybeans/injection/EntityManagerHelper",
                "getEntityManager",
                "(Lorg/objectweb/easybeans/container/session/EasyBeansSessionContext;"
                + "Ljava/lang/String;Ljavax/persistence/PersistenceContextType;)"
                + "Ljavax/persistence/EntityManager;");
    }

    /**
     * Add a call to EntityManagerHelper class (PersistenceUnit):
     * EntityManagerHelper.getEntityManagerFactory(getEasyBeansSessionContext(), unitName).
     * @param javaxPersistenceUnit informations on the persistence unit
     * @param mv the method visitor on which instructions are added
     */
    private void addCallEntityManagerFactoryHelper(final JavaxPersistenceUnit javaxPersistenceUnit, final MethodVisitor mv) {
        // get EasyBeansSessionContext
        addCallGetEasyBeansContext(mv);

        // Persistence-unit name
        mv.visitLdcInsn(javaxPersistenceUnit.getUnitName());

        mv.visitMethodInsn(INVOKESTATIC, "org/objectweb/easybeans/injection/EntityManagerHelper",
                "getEntityManagerFactory",
                "(Lorg/objectweb/easybeans/container/session/EasyBeansSessionContext;"
                + "Ljava/lang/String;)Ljavax/persistence/EntityManagerFactory;");
    }


    /**
     * Generates a call to JNDILookupHelper class to get the java:comp/env name
     * requested.
     * @param jndiName the name to lookup.
     * @param type the ASM type
     * @param mv the method visitor to write code.
     * @param className the name of the generated class.
     * @param jndiType the type of access (registry, java:comp/env, etc)
     */
    private void callJndi(final String jndiName, final Type type, final MethodVisitor mv,
            final String className, final JndiType jndiType) {
        if (!staticMode) {
            mv.visitVarInsn(ALOAD, 0);
        }
        mv.visitLdcInsn(jndiName);
        String mName = "";
        switch (jndiType) {
            case JAVA_COMP:
                mName = "getCompJndiName";
                break;
            case JAVA_COMP_ENV:
                mName = "getEnvJndiName";
                break;
            case REGISTRY:
                mName = "getJndiName";
                break;
            default:
                throw new IllegalStateException("invalid type");
        }
        mv.visitMethodInsn(INVOKESTATIC, "org/objectweb/easybeans/injection/JNDILookupHelper", mName,
                "(Ljava/lang/String;)Ljava/lang/Object;");
        CommonClassGenerator.transformObjectIntoPrimitive(type, mv);
    }

    /**
     * Generates a call to JNDILookupHelper class to get the java:comp/env name
     * requested.
     * @param jndiName the name to lookup.
     * @param type the ASM type.
     * @param mv the method visitor to write code.
     * @param fieldMetaData the metadata of the attribute.
     * @param className the name of the generated class.
     * @param jndiType the type of access (registry, java:comp/env, etc)
     */
    private void callAttributeJndi(final String jndiName, final Type type, final MethodVisitor mv,
            final FieldAnnotationMetadata fieldMetaData, final String className, final JndiType jndiType) {
        logger.debug("Add injection for @Resource on attribute {0} of class {1} for the type {2}", fieldMetaData
                .getFieldName(), className, type.getClassName());

        String formattedJndiName = getJndiName(jndiName, fieldMetaData);
        callJndi(formattedJndiName, type, mv, className, jndiType);
        setField(mv, className, fieldMetaData, type);
    }

    /**
     * Sets the value of a field.
     * @param mv the method visitor to use
     * @param className the class which contains the attribute.
     * @param fieldMetaData the metadata corresponding to the attribute
     * @param type the ASM type of the attribute
     */
    private void setField(final MethodVisitor mv, final String className,
            final FieldAnnotationMetadata fieldMetaData, final Type type) {
        mv.visitFieldInsn(setField(), className, fieldMetaData.getFieldName(), type.getDescriptor());
    }

    /**
     * Sets the value of a field.
     * @param mv the method visitor to use
     * @param className the class which contains the method.
     * @param methodMetaData the metadata corresponding to the method *
     * @param type the ASM type of the method
     */
    private void callSetterMethod(final MethodVisitor mv, final String className,
            final MethodAnnotationMetadata methodMetaData, final Type type) {
        mv.visitMethodInsn(INVOKEVIRTUAL, className, methodMetaData.getMethodName(), methodMetaData.getJMethod()
                .getDescriptor());
    }


    /**
     * Generates a call to JNDILookupHelper class to get the java:comp/env name
     * requested.
     * @param jndiName the name to lookup.
     * @param type the ASM type.
     * @param mv the method visitor to write code.
     * @param methodMetaData the metadata of the method.
     * @param className the name of the generated class.
     * @param jndiType the type of access (registry, java:comp/env, etc)
     */
    private void callMethodJndiEnv(final String jndiName, final Type type, final MethodVisitor mv,
            final MethodAnnotationMetadata methodMetaData, final String className, final JndiType jndiType) {
        logger.debug("Add injection for @Resource on method {0} of class {1} for the type {2}", methodMetaData
                .getMethodName(), className, type.getClassName());

        String checkedJndiName = getJndiName(jndiName, methodMetaData);
        callJndi(checkedJndiName, type, mv, className, jndiType);
        callSetterMethod(mv, className, methodMetaData, type);
    }


    /**
     * Ensures that jndiName is valid and return the default value if it is a null value.
     * @param jndiName to check
     * @param fieldAnnotationMetadata attribute's metadata
     * @return jndi name value.
     */
    private String getJndiName(final String jndiName, final FieldAnnotationMetadata fieldAnnotationMetadata) {
        String newJndiName = jndiName;
        if (jndiName == null || "".equals(jndiName)) {
            logger.debug("Name property undefined.");
            newJndiName = classAnnotationMetadata.getClassName().replace("/", ".") + "/" + fieldAnnotationMetadata.getFieldName();
            logger.debug("Getting environment's entry with default JNDI name: {0}",  newJndiName);
        }
        return newJndiName;
    }


    /**
     * Ensures that jndiName is valid and return the default value if it is a null value.
     * @param jndiName to check
     * @param methodMetaData attribute's metadata
     * @return jndi name value.
     */
    private String getJndiName(final String jndiName, final MethodAnnotationMetadata methodMetaData) {
        String newJndiName = jndiName;
        if (jndiName == null || "".equals(jndiName)) {
            logger.debug("Property name not defined.");
            StringBuilder propertyBuilder = new StringBuilder(methodMetaData.getMethodName());
            propertyBuilder.delete(0, LENGTH);
            propertyBuilder.setCharAt(0, Character.toLowerCase(propertyBuilder.charAt(0)));
            propertyBuilder.insert(0, classAnnotationMetadata.getClassName().replace("/", ".") + "/");
            newJndiName = propertyBuilder.toString();
            logger.debug("Getting environment's entry with default JNDI name: {0}",  newJndiName);
        }

        return newJndiName;
    }


    /**
     * Generates a call to JNDIBinderHelper class to bind an object into the
     * java:comp/env context.
     * @param encName the name to bind
     * @param jndiName the name to link to (LinkRef)
     * @param mv the method visitor to write code.
     * @param fieldMetaData the metadata of the attribute (that will be bound)
     */
    private void callBindAttributeJndi(final String encName, final String jndiName, final MethodVisitor mv,
            final FieldAnnotationMetadata fieldMetaData) {


        // call : JNDIBinderHelper.bindLinkRefEnvJndiName("localName", "jndiName");
        mv.visitLdcInsn(getJndiName(encName, fieldMetaData));
        mv.visitLdcInsn(jndiName);
        mv.visitMethodInsn(INVOKESTATIC, "org/objectweb/easybeans/injection/JNDIBinderHelper",
                "bindLinkRefEnvJndiName", "(Ljava/lang/String;Ljava/lang/String;)V");
    }


    /**
     * Generates a call to JNDIBinderHelper class to link an object into the
     * java:comp/env context.
     * @param encName the name to bind
     * @param jndiName the jndi name to use for the link
     * @param mv the method visitor to write code.
     */
    private void callBindLookupJndiRef(final String encName, final String jndiName, final MethodVisitor mv) {
        // use a LinkRef between JNDI name and ENC name
        // call : JNDIBinderHelper.bindLinkRefEnvJndiName("localName", "jndiName");

        mv.visitLdcInsn(encName);
        mv.visitLdcInsn(jndiName);
        mv.visitMethodInsn(INVOKESTATIC, "org/objectweb/easybeans/injection/JNDIBinderHelper",
                "bindLinkRefEnvJndiName", "(Ljava/lang/String;Ljava/lang/String;)V");

        logger.debug("Linking Object with JNDI name '" + jndiName + "' to ENC name '" + encName + "' for the class '"
                + classAnnotationMetadata.getClassName() + "'.");
    }


    /**
     * Bind a ref for an EJB in ENC environment.
     * @param jEJB annotation to analyze
     * @param mv the visitor on which write the bytecode.
     */
    private void bindClassEJB(final JEjbEJB jEJB, final MethodVisitor mv) {

        // JNDIBinderHelper.bindEnvJndiName("encName",
        // JNDILookupHelper.getEnvJndiName("jndiName"));


        // name attribute is mandatory
        String encName = jEJB.getName();
        if (encName == null || "".equals(encName)) {
            throw new IllegalStateException("Error when analyzing @EJB annotation '" + jEJB
                    + "' for the class '" + classAnnotationMetadata.getClassName() + "' : No name !");
        }

        // jndiName
        JNDIResolver jndiResolver = (JNDIResolver) map.get(JNDIResolver.NAME);
        // ejbName ?
        String beanName = jEJB.getBeanName();
        String jndiName = null;
        if (jEJB.getMappedName() != null && jEJB.getMappedName().length() > 0) {
            jndiName = jEJB.getMappedName();
        } else {
            jndiName = jndiResolver.getJndiNameInterface(jEJB.getBeanInterface(), beanName);
        }

        if (jndiName == null) {
            throw new IllegalStateException("No JNDI name found when analyzing @EJB annotation '" + jEJB
                    + "' for the class '" + classAnnotationMetadata.getClassName() + "'.");
        }

        // inject code
        callBindLookupJndiRef(encName, jndiName, mv);
    }

    /**
     * Bind a ref for a Resource in ENC environment.
     * @param jAnnotationResource annotation to analyze
     * @param mv the visitor on which write the bytecode.
     */
    private void bindResource(final JAnnotationResource jAnnotationResource, final MethodVisitor mv) {
        if (DATASOURCE_ITF.equals(jAnnotationResource.getType())
                || SESSION_CONTEXT.equals(jAnnotationResource.getType())) {
            // get name
            String encName = jAnnotationResource.getName();
            if (encName == null || "".equals(encName)) {
                logger.error("No encName for Annotation resource {0}.", jAnnotationResource);
                return;
            }
            String jndiName = null;
            if (SESSION_CONTEXT.equals(jAnnotationResource.getType())) {
                jndiName = "java:comp/EJBContext";
            } else {
                jndiName = jAnnotationResource.getMappedName();
            }
            if (jndiName == null) {
                logger.error("JndiName for resource annotation {0} is null, no binding to ENC name {1}",
                        jAnnotationResource, encName);
            } else {
                // inject code
                callBindLookupJndiRef(encName, jndiName, mv);
            }
        }
    }


    /**
     * Bind a ref for a PersistenceContext in ENC environment.
     * @param javaxPersistenceContext annotation to analyze
     * @param mv the visitor on which write the bytecode.
     */
    private void bindClassPersistenceContext(final JavaxPersistenceContext javaxPersistenceContext, final MethodVisitor mv) {

        // Add :
        // JNDIBinderHelper.bindEnvJndiName("myName", EntityManagerHelper.getEnt....);

        // name for ENC
        mv.visitLdcInsn(javaxPersistenceContext.getName());

        // call em helper
        addCallEntityManagerHelper(javaxPersistenceContext, mv);

        // bind in ENC
        mv.visitMethodInsn(INVOKESTATIC, "org/objectweb/easybeans/injection/JNDIBinderHelper", "bindEnvJndiName",
        "(Ljava/lang/String;Ljava/lang/Object;)V");
    }


    /**
     * Bind a ref for a PersistenceUnit in ENC environment.
     * @param javaxPersistenceUnit annotation to analyze
     * @param mv the visitor on which write the bytecode.
     */
    private void bindClassPersistenceUnit(final JavaxPersistenceUnit javaxPersistenceUnit, final MethodVisitor mv) {
        // Add :
        // JNDIBinderHelper.bindEnvJndiName("myName", EntityManagerFactory.getEnt....);

        // name for ENC
        mv.visitLdcInsn(javaxPersistenceUnit.getName());

        // get EMF
        addCallEntityManagerFactoryHelper(javaxPersistenceUnit, mv);

        // bind in ENC
        mv.visitMethodInsn(INVOKESTATIC, "org/objectweb/easybeans/injection/JNDIBinderHelper", "bindEnvJndiName",
        "(Ljava/lang/String;Ljava/lang/Object;)V");
    }


    /**
     * @return the opCode used for non-static or static mode.
     */
    private int setField() {
        int opCode = PUTFIELD;
        if (staticMode) {
            opCode = PUTSTATIC;
        }
        return opCode;
    }

    /**
     * Check that the annotation is not used on a final field or static field
     * (except client mode).
     * @param field the attribute to check.
     */
    private void validateAccessFieldAnnotation(final FieldAnnotationMetadata field) {
        if (accessTest(field.getJField().getAccess(), ACC_FINAL)) {
            throw new IllegalStateException("The '" + field
                    + "' attribute is a final attribute which is not compliant for dependency injection.");
        }
        if (!staticMode && accessTest(field.getJField().getAccess(), ACC_STATIC)) {
            throw new IllegalStateException("The '" + field
                    + "' attribute is a static attribute which is not compliant for dependency injection.");
        }
    }

    /**
     * Check that the annotation is not used on a final field or static field
     * (except client mode).
     * @param methodData the method to check.
     */
    private void validateAccessMethodAnnotation(final MethodAnnotationMetadata methodData) {
        if (!staticMode && accessTest(methodData.getJMethod().getAccess(), ACC_STATIC)) {
            throw new IllegalStateException("The '" + methodData
                    + "' method is a static attribute which is not compliant for dependency injection.");
        }
    }


    /**
     * @param access the full access modified to check
     * @param checkedAccess the access to check
     * @return true if it is ok, else false.
     */
    private boolean accessTest(final int access, final int checkedAccess) {
        return (access & checkedAccess) == checkedAccess;
    }
}
