/**
 * EasyBeans
 * Copyright (C) 2006 Bull S.A.S.
 * Contact: easybeans@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: MBeansHelper.java 655 2006-06-14 09:40:48Z sauthieg $
 * --------------------------------------------------------------------------
 */

package org.objectweb.easybeans.jmx;

import java.util.Hashtable;
import java.util.Map;

import javax.management.MalformedObjectNameException;
import javax.management.ObjectName;

import org.objectweb.easybeans.api.jmx.EZBManagementIdentifier;
import org.objectweb.easybeans.log.JLog;
import org.objectweb.easybeans.log.JLogFactory;

/**
 * Singleton object.Creates the MBeans and register them.
 * @author florent
 *
 */
public final class MBeansHelper {
    /**
     * Logger.
     */
    private static JLog logger = JLogFactory.getLog(MBeansHelper.class);

    /**
     * Singleton instance.
     */
    private static MBeansHelper instance = null;

    /**
     * The Identifier in charge of creating the right ObjectName for a given
     * instance.
     */
    private Map<Class, EZBManagementIdentifier> identifiers = null;

    /**
     * Singleton class, no public constructor.
     */
    private MBeansHelper() {
        identifiers = new Hashtable<Class, EZBManagementIdentifier>();
    }

    /**
     * @return Returns the Singleton MBeansHelper instance.
     */
    public static MBeansHelper getInstance() {
        if (instance == null) {
            instance = new MBeansHelper();
        }
        return instance;
    }

    /**
     * Register the instance as a ModelMBean using the delegate.
     * @param <T> instance Type
     * @param instance Object instance to be managed
     * @throws MBeansException if registration fails.
     */
    public <T> void registerMBean(final T instance) throws MBeansException {

        String on = getObjectName(instance);

        // register
        try {
            CommonsModelerHelper.registerModelerMBean(instance, on);
        } catch (CommonsModelerException e) {
            throw new MBeansException("Cannot register MBean", e);
        }
    }

    /**
     * Unregister the given Object.
     * @param <T> instance Type
     * @param instance Instance to be deregistered.
     * @throws MBeansException if unregistration fails.
     */
    public <T> void unregisterMBean(final T instance) throws MBeansException {

        String on = getObjectName(instance);

        // register
        try {
            CommonsModelerHelper.unregisterModelerMBean(new ObjectName(on));
        } catch (MalformedObjectNameException e) {
            logger.warn("Cannot unregister MBean '" + on + "' : " + e.getMessage());
        } catch (NullPointerException e) {
            logger.warn("Cannot unregister MBean '" + on + "' : " + e.getMessage());
        }
    }

    /**
     * @param <T> instance Type
     * @param instance Object instance to be managed
     * @return Returns the instance ObjectName.
     * @throws MBeansException if registration fails.
     */
    public <T> String getObjectName(final T instance) throws MBeansException {

        EZBManagementIdentifier<T> identifier = getIdentifier(instance);

        // gather the ObjectName
        if (identifier != null) {
            StringBuilder sb = new StringBuilder();

            sb.append(identifier.getDomain());
            sb.append(":");
            sb.append(identifier.getTypeProperty());
            sb.append(",");
            sb.append("name=");
            sb.append(identifier.getNamePropertyValue(instance));
            String additionnal = identifier.getAdditionnalProperties(instance);
            if (additionnal != null && (!"".equals(additionnal))) {
                sb.append(",");
                sb.append(additionnal);
            }
            return sb.toString();
        }

        return null;

    }

    /**
     * @param <T> instance type
     * @param instance instance to be managed.
     * @return Returns an {@link EZBManagementIdentifier} for the given Resource type.
     * @throws MBeansException if the Identifier cannot be returned.
     */
    @SuppressWarnings("unchecked")
    private <T> EZBManagementIdentifier<T> getIdentifier(final T instance) throws MBeansException {

        // try to use the cached identifier
        Class clz = instance.getClass();
        if (identifiers.containsKey(clz)) {
            return identifiers.get(clz);
        }

        // the identifier was not cached
        String mbeanClassname = instance.getClass().getName();
        String mbeanPackage = mbeanClassname.substring(0, mbeanClassname
                .lastIndexOf(".") + 1);

        // looking for a class named :
        // <mbean-package>.management.<mbean-model-name>Identifier
        // ex : org.objectweb.easybeans.container.management.JContainer3Identifier
        String identifierClassname = mbeanPackage + "management." + clz.getSimpleName() + "Identifier";

        // Instantiate it ...
        ClassLoader loader = Thread.currentThread().getContextClassLoader();
        try {
            Class<? extends EZBManagementIdentifier> cls = loader.loadClass(
                    identifierClassname).asSubclass(EZBManagementIdentifier.class);
            EZBManagementIdentifier<T> id = cls.newInstance();

            // cache the identifier
            identifiers.put(clz, id);

            return id;
        } catch (ClassNotFoundException e) {
            throw new MBeansException("Identifier Class not found", e);
        } catch (InstantiationException e) {
            throw new MBeansException("Identifier Class not instantiated", e);
        } catch (IllegalAccessException e) {
            throw new MBeansException("Identifier Class not instantiated", e);
        }
    }

}
